from typing import Optional, List, Dict, Any
import uuid

from dataclasses import dataclass

from .base_codecs import Codec
from .identity_codecs import Identity


@dataclass
class Detection(Codec):
    """An object detected in a video frame. Detections can have attributes
    attached to them that provide more information about the object as well as
    other metadata like a unique tracking ID.
    """

    class_name: str
    """The class of object that was detected, like 'person' or 'car'"""

    coords: List[List[int]]
    """The coordinates, in pixels, of the detection in the frame"""

    children: List["Detection"]

    attributes: Dict[str, str]
    """A dict whose key is an attribute name and whose value is the value
    of that attribute. For example, a car detection may have an attribute
    whose key is 'type' and whose value is 'sedan'.
    """

    with_identity: Optional[Identity]
    """If not None, this is the identity that this detection was recognized as.
    """

    extra_data: Dict[str, Any]
    """Any additional metadata describing this object"""

    track_id: Optional[uuid.UUID]
    """If not None, this is a unique tracking ID for the object. This ID
    can be compared to detections from other frames to track the movement
    of an object over time.
    """

    @property
    def center(self):
        """Return the center of the detections coordinates"""
        x = [c[0] for c in self.coords]
        y = [c[1] for c in self.coords]
        return sum(x) / len(x), sum(y) / len(y)

    @property
    def bbox(self):
        sorted_x = sorted([c[0] for c in self.coords])
        sorted_y = sorted([c[1] for c in self.coords])
        return [[sorted_x[0], sorted_y[0]],
                [sorted_x[-1], sorted_y[0]],
                [sorted_x[-1], sorted_y[-1]],
                [sorted_x[0], sorted_y[-1]]]

    def to_dict(self):
        d = dict(self.__dict__)
        if self.with_identity:
            d["with_identity"] = Identity.to_dict(d["with_identity"])
        if self.track_id:
            d["track_id"] = str(self.track_id)

        d["children"] = [Detection.to_dict(det) for det in self.children]
        return d

    @staticmethod
    def from_dict(d):
        with_identity = None
        if d["with_identity"]:
            with_identity = Identity.from_dict(d["with_identity"])

        track_id = None
        if d["track_id"]:
            track_id = uuid.UUID(d["track_id"])

        children = [Detection.from_dict(det) for det in d["children"]]
        return Detection(class_name=d["class_name"],
                         coords=d["coords"],
                         children=children,
                         attributes=d["attributes"],
                         with_identity=with_identity,
                         extra_data=d["extra_data"],
                         track_id=track_id)


@dataclass
class Attribute(Codec):
    """This holds an attribute of a detection. These should _not_ be made
    on the client side
    """

    category: str
    """The category of attribute being described"""

    value: str
    """The value for this attribute category"""

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_dict(d):
        return Attribute(category=d["category"],
                         value=d["value"])



from typing import Dict, List, Optional

import json

from brainframe.api.bf_codecs import Plugin
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class PluginStubMixin(BaseStub):
    """Provides stubs to call APIs to inspect and configure plugins."""

    def get_plugin(self, name,
                   timeout=DEFAULT_TIMEOUT) -> Plugin:
        """
        :param name: The name of the plugin to get
        :param timeout: The timeout to use for this request
        :return: Plugin with the given name
        """
        req = f"/api/plugins/{name}"
        plugin, _ = self._get_json(req, timeout)
        return Plugin.from_dict(plugin)

    def get_plugins(self, timeout=DEFAULT_TIMEOUT) -> List[Plugin]:
        """
        :param timeout: The timeout to use for this request
        :return: All available plugins
        """
        req = "/api/plugins"
        plugins, _ = self._get_json(req, timeout)
        return [Plugin.from_dict(d) for d in plugins]

    def get_plugin_option_vals(self, plugin_name, stream_id=None,
                               timeout=DEFAULT_TIMEOUT) \
            -> Dict[str, object]:
        """Gets the current values for every plugin option. See the
        documentation for the PluginOption codec for more info about global and
        stream level options and how they interact.

        :param plugin_name: The plugin to find options for
        :param stream_id: The ID of the stream. If this value is None, then the
            global options are returned for that plugin
        :param timeout: The timeout to use for this request
        :return: A dict where the key is the option name and the value is the
            option's current value
        """
        if stream_id is None:
            req = f"/api/plugins/{plugin_name}/options"
        else:
            req = f"/api/streams/{stream_id}/plugins/{plugin_name}/options"
        plugin_option_vals, _ = self._get_json(req, timeout)

        return plugin_option_vals

    def set_plugin_option_vals(self, *, plugin_name, stream_id=None,
                               option_vals: Dict[str, object],
                               timeout=DEFAULT_TIMEOUT):
        """Sets option values for a plugin.

        :param plugin_name: The name of the plugin whose options to set
        :param stream_id: The ID of the stream, if these are stream-level
            options. If this value is None, then the global options are set
        :param option_vals: A dict where the key is the name of the option to
            set, and the value is the value to set that option to
        :param timeout: The timeout to use for this request
        """
        if stream_id is None:
            req = f"/api/plugins/{plugin_name}/options"
        else:
            req = f"/api/streams/{stream_id}/plugins/{plugin_name}/options"

        option_values_json = json.dumps(option_vals)
        self._put_json(req, timeout, option_values_json)

    def patch_plugin_option_vals(self, *, plugin_name, stream_id=None,
                                 option_vals: Dict[str, object],
                                 timeout=DEFAULT_TIMEOUT):
        """Patches option values for a plugin. Only the provided options are
        changed. To unset an option, provide that option with a value of None.

        :param plugin_name: The name of the plugin whose options to set
        :param stream_id: The ID of the stream, if these are stream-level
            options. If this value is None, then the global options are set
        :param option_vals: A dict where the key is the name of the option to
            set, and the value is the value to set that option to
        :param timeout: The timeout to use for this request
        """
        if stream_id is None:
            req = f"/api/plugins/{plugin_name}/options"
        else:
            req = f"/api/streams/{stream_id}/plugins/{plugin_name}/options"

        option_values_json = json.dumps(option_vals)
        self._patch_json(req, timeout, option_values_json)

    def is_plugin_active(self, plugin_name, stream_id=None,
                         timeout=DEFAULT_TIMEOUT) -> bool:
        """Returns True if the plugin is active. If a plugin is not marked as
        active, it will not run. Like plugin options, this can be configured
        globally and on a per-stream level.

        :param plugin_name: The name of the plugin to get activity for
        :param stream_id: The ID of the stream, if you want the per-stream
            active setting
        :param timeout: The timeout to use for this request
        :return: True if the plugin is active
        """
        if stream_id is None:
            req = f"/api/plugins/{plugin_name}/active"
        else:
            req = f"/api/streams/{stream_id}/plugins/{plugin_name}/active"
        plugins_active, _ = self._get_json(req, timeout)
        return plugins_active

    def set_plugin_active(self, *, plugin_name, stream_id=None,
                          active: Optional[bool],
                          timeout=DEFAULT_TIMEOUT):
        """Sets whether or not the plugin is active. If a plugin is active, it
        will be run on frames.

        :param plugin_name: The name of the plugin to set activity for
        :param stream_id: The ID of the stream, if you want to set the
            per-stream active setting
        :param active: True if the plugin should be set to active
        :param timeout: The timeout to use for this request
        """
        if stream_id is None:
            req = f"/api/plugins/{plugin_name}/active"
        else:
            req = f"/api/streams/{stream_id}/plugins/{plugin_name}/active"

        active_json = json.dumps(active)

        self._put_json(req, timeout, active_json)



from .alerts import AlertStubMixin
from .analysis import AnalysisStubMixin
from .identities import IdentityStubMixin
from .plugins import PluginStubMixin
from .streams import StreamStubMixin
from .zone_statuses import ZoneStatusStubMixin
from .zones import ZoneStubMixin
from .storage import StorageStubMixin
from .alarms import ZoneAlarmStubMixin
from .process_image import ProcessImageStubMixIn
from .encodings import EncodingStubMixIn
from .premises import PremisesStubMixin
from .users import UserStubMixin
from .licenses import LicenseStubMixIn
from .base_stub import BaseStub



from typing import Dict, Generator

import json

from brainframe.api.bf_codecs import ZoneStatus
from .base_stub import BaseStub, DEFAULT_TIMEOUT

ZONE_STATUS_TYPE = Dict[int, Dict[str, ZoneStatus]]
ZONE_STATUS_STREAM_TYPE = Generator[ZONE_STATUS_TYPE, None, None]


class ZoneStatusStubMixin(BaseStub):
    """Provides stubs for calling APIs to get zone statuses."""

    def get_latest_zone_statuses(self,
                                 timeout=DEFAULT_TIMEOUT) -> ZONE_STATUS_TYPE:
        """This method gets all of the latest processed zone statuses for every
        zone and for every stream.

        All active streams will have a key in the output dict.

        :param timeout: The timeout to use for this request
        :return: A dict whose keys are stream IDs and whose value is another
            dict. This nested dict's keys are zone names and their value is the
            ZoneStatus for that zone.
        """
        req = "/api/streams/status"
        data, _ = self._get_json(req, timeout)

        # Convert ZoneStatuses to Codecs
        out = {int(s_id): {key: ZoneStatus.from_dict(val)
                           for key, val in statuses.items()}
               for s_id, statuses in data.items()}
        return out

    def get_zone_status_stream(self, timeout=None) -> ZONE_STATUS_STREAM_TYPE:
        """Streams ZoneStatus results from the server as they are produced.

        All active streams will have a key in the output dict.

        :param timeout: The timeout to use for this request
        :return: A generator that outputs dicts whose keys are stream IDs and
            whose value is another dict. This nested dict's keys are zone names
            and their value is the ZoneStatus for that zone.
        """
        req = "/api/streams/statuses"

        def zone_status_iterator():
            # Don't use a timeout for this request, since it's ongoing
            resp = self._get(req, timeout=timeout)
            for packet in resp.iter_lines(delimiter=b"\r\n"):
                if packet == b'':
                    continue

                # Parse the line
                zone_statuses_dict = json.loads(packet)

                processed = {int(s_id): {key: ZoneStatus.from_dict(val)
                                         for key, val in statuses.items()}
                             for s_id, statuses in zone_statuses_dict.items()}
                yield processed

        return zone_status_iterator()



from enum import Enum


class Ordering(Enum):
    """Specifies in what order a field should be sorted by."""
    ASC = "asc"
    """The order of the field should be ascending (from low to high)"""
    DESC = "desc"
    """The order of the field should be descending (from high to low)"""


class SortOptions:
    """A sorting configuration. Used by some APIs that provide many of a
    certain object.
    """

    def __init__(self, field_name: str, ordering: Ordering):
        self.field_name: str = field_name
        """The name of the field to sort by"""
        self.ordering: Ordering = ordering
        """The order to sort the field by"""

    def query_format(self) -> str:
        return f"{self.field_name}:{self.ordering.value}"



from typing import Optional
from enum import Enum

from dataclasses import dataclass

from .base_codecs import Codec


class RoleType(Enum):
    """Controls the level of access a user has to API endpoints."""

    EDITOR = "editor"
    """A user that can access most endpoints but cannot do administrative tasks
    like adding users and managing the license.
    """
    ADMIN = "admin"
    """A user that can access all endpoints."""

    @classmethod
    def values(cls):
        return [v.value for v in cls]


@dataclass
class User(Codec):
    """Contains information on a user."""

    username: str
    """The username used for login"""

    password: Optional[str]
    """This field will be None when retrieving users from the server. It
    should only be set by the client when creating a new user or updating a
    user's password.
    """

    role: RoleType
    """The user's role"""

    id: Optional[int] = None
    """The user's unique ID"""

    def to_dict(self) -> dict:
        d = dict(self.__dict__)
        d["role"] = self.role.value
        return d

    @staticmethod
    def from_dict(d: dict):
        return User(
            username=d["username"],
            password=d["password"],
            role=RoleType(d["role"]),
            id=d["id"],
        )



import abc
import json


class Codec(abc.ABC):
    """A serializable object."""

    @abc.abstractmethod
    def to_dict(self) -> dict:
        pass

    @abc.abstractstaticmethod
    def from_dict(d: dict):
        pass

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, j: str):
        return cls.from_dict(json.loads(j))

    def __eq__(self, other):
        if type(other) is dict:
            return self.to_dict() == other

        if type(other) is not type(self):
            return NotImplemented

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        if type(other) is not type(self):
            return NotImplemented

        return self.to_dict() != other.to_dict()



from typing import Optional

from dataclasses import dataclass

from .base_codecs import Codec


@dataclass
class Premises(Codec):
    """Information about a specific Premises"""

    name: str
    """The friendly name of the premises"""

    id: Optional[int] = None
    """The unique ID of the premises"""

    def to_dict(self):
        d = dict(self.__dict__)
        return d

    @staticmethod
    def from_dict(d):
        return Premises(name=d["name"], id=d["id"])



from enum import Flag, auto
from typing import List, Optional, Tuple

import numpy as np
import json

from brainframe.api.bf_errors import FrameNotFoundForAlertError
from brainframe.api.bf_codecs import Alert, image_utils
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class AlertStubMixin(BaseStub):
    """Provides stubs for calling APIs related to controlling and getting
    alerts.
    """

    class AlertVerificationQueryType(Flag):
        UNVERIFIED = auto()
        TRUE = auto()
        FALSE = auto()
        VERIFIED = TRUE | FALSE
        NO_FALSE_POSITIVES = UNVERIFIED | TRUE
        NO_TRUE_POSITIVES = UNVERIFIED | FALSE
        ALL = UNVERIFIED | TRUE | FALSE

        @property
        def query_repr(self):
            values = []
            if self & self.UNVERIFIED:
                values.append("unverified")
            if self & self.TRUE:
                values.append("true")
            if self & self.FALSE:
                values.append("false")
            return ','.join(values)

    def get_alert(self, alert_id, timeout=DEFAULT_TIMEOUT) -> Alert:
        """Gets the alert with the given ID.

        :param alert_id: The ID of the alert to get
        :param timeout: The timeout to use for this request
        :return: The alert
        """
        req = f"/api/alerts/{alert_id}"
        data, _ = self._get_json(req, timeout)

        return Alert.from_dict(data)

    def get_alerts(self,
                   stream_id: Optional[int] = None,
                   zone_id: Optional[int] = None,
                   alarm_id: Optional[int] = None,
                   verification: Optional[AlertVerificationQueryType]
                   = AlertVerificationQueryType.ALL,
                   limit: Optional[int] = None,
                   offset: Optional[int] = None,
                   timeout=DEFAULT_TIMEOUT) \
            -> Tuple[List[Alert], int]:
        """Gets all alerts that match a query

        :param stream_id: The ID of the stream to get alerts for
        :param zone_id: The ID of the zone to get alerts for
        :param alarm_id: The ID of the alarm to get alerts for
        :param verification: The verification states of the alerts
        :param limit: The maximum number of alerts to return. If None, no limit
            will be applied
        :param offset: The offset from the most recent alerts to return. This
            is only useful when providing a limit.
        :param timeout: The timeout to use for this request
        :return:  A list of alerts, and the total number of alerts that
            fit this criteria, ignoring pagination (the limit and offset)
        """
        req = "/api/alerts"

        params = {}
        if stream_id is not None:
            params["stream_id"] = stream_id
        if zone_id is not None:
            params["zone_id"] = zone_id
        if alarm_id is not None:
            params["alarm_id"] = alarm_id
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if verification is not self.AlertVerificationQueryType.ALL:
            params["verification"] = verification.query_repr

        data, headers = self._get_json(req, timeout, params=params)
        alerts = [Alert.from_dict(a) for a in data]

        total_count = int(headers["Total-Count"])

        return alerts, total_count

    def set_alert_verification(self, alert_id, verified_as: bool,
                               timeout=DEFAULT_TIMEOUT):
        """Sets an alert verified as True or False.

        :param alert_id: The ID of the alert to set
        :param verified_as: Set verification to True or False
        :param timeout: The timeout to use for this request
        :return: The modified Alert
        """
        req = f"/api/alerts/{alert_id}"
        self._put_json(req, timeout, json.dumps(verified_as))

    def get_alert_frame(self, alert_id: int,
                        timeout=DEFAULT_TIMEOUT) -> Optional[np.ndarray]:
        """Returns the frame saved for this alert, or None if no frame is
        recorded for this alert.

        :param alert_id: The ID of the alert to get a frame for.
        :param timeout: The timeout to use for this request
        :return: The image as loaded by OpenCV, or None
        """
        req = f"/api/alerts/{alert_id}/frame"
        try:
            img_bytes = self._get(req, timeout).content
            return image_utils.decode(img_bytes)
        except FrameNotFoundForAlertError:
            return None



from .base_codecs import Codec

from .alarm_codecs import Alert, ZoneAlarm
from .condition_codecs import (
    ZoneAlarmCountCondition,
    ZoneAlarmRateCondition,
    IntersectionPointType,
    CountConditionTestType,
    RateConditionTestType,
    DirectionType,
)
from .config_codecs import StreamConfiguration, ConnType
from .identity_codecs import Identity, Encoding
from .detection_codecs import Attribute, Detection
from .zone_codecs import Zone, ZoneStatus
from .plugin_codecs import (
    PluginOption,
    Plugin,
    NodeDescription,
    OptionType,
    SizeType,
)
from .premises_codecs import Premises
from .user_codecs import User, RoleType
from .license_codecs import (
    LicenseTerms,
    LicenseInfo,
    DATE_FORMAT,
    LicenseState,
)
from .sorting import SortOptions, Ordering



import logging
from typing import Any, BinaryIO, Optional, Tuple, Union
from urllib.parse import urlparse

import requests
import json
from requests import Response

from brainframe.api.bf_codecs import Codec
from brainframe.api.bf_errors import InvalidSessionError, UnknownError, \
    kind_to_error_type

DEFAULT_TIMEOUT = 30
"""The default timeout for most requests."""


class BaseStub:
    """A base class for API stubs: Classes that provide methods which call the
    API to provide some functionality. Included are methods to make
    communicating with the API easier.
    """

    _server_url = None
    _credentials = None
    _session_id = None

    def set_url(self, url):
        scheme = urlparse(url).scheme
        if scheme not in ["http", "https"]:
            raise ValueError(f"Invalid URL schema ({scheme}://). Must be "
                             f"either http:// or https://")

        self._server_url = url

    def set_credentials(self, credentials: Optional[Tuple[str, str]]):
        """Start authorizing requests with the given credentials for all future
        requests.

        :param credentials: The username and password in a tuple, or None to
            not use authorization on requests
        """
        self._credentials = credentials

        # Stop using the old session with outdated credentials
        self._session_id = None

    def _get_json(self, api_url, timeout, params=None) -> Tuple[Any, dict]:
        """Send a GET request to the given URL and parse the result as JSON.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param params: The "query_string" to add to the url. In the format
            of a dict, {"key": "value", ...} key and val must be a string
        :return: The response, parsed with JSON
        """
        resp = self._get(api_url, timeout, params=params)

        if resp.content:
            return json.loads(resp.content), resp.headers
        return None, resp.headers

    def _put_json(self, api_url, timeout, json_data):
        """Send a PUT request to the given URL.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param json_data: Pre-formatted JSON to send
        :return: The JSON response as a dict, or None if none was sent
        """
        resp = self._put(api_url,
                         timeout,
                         data=json_data,
                         content_type="application/json")

        if resp.content:
            return json.loads(resp.content)
        return None

    def _post_codec(self, api_url, timeout, codec: Codec):
        """Send a POST request to the given URL.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param codec: A codec to convert to JSON and send
        :return: The JSON response as a dict, or None if none was sent
        """
        codec_data = codec.to_json()
        resp = self._post(api_url,
                          timeout,
                          data=codec_data.encode("utf-8"),
                          content_type="application/json")

        if resp.content:
            return json.loads(resp.content)
        return None

    def _post_json(self, api_url, timeout, json_data):
        """Send a POST request to the given URL.
        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param json_data: Pre-formatted JSON to send
        :return: The JSON response as a dict, or None if none was sent
        """
        resp = self._post(api_url,
                          timeout,
                          data=json_data,
                          content_type="application/json")

        if resp.content:
            return json.loads(resp.content)
        return None

    def _post_multipart(self, api_url, timeout, files):
        """Send a POST request to the given URL.
        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param files: A tuple in Requests format for a multipart body
        :return: The JSON response as a dict, or None if none was sent
        """
        resp = self._post(api_url, timeout, files=files)

        if resp.content:
            return json.loads(resp.content)
        return None

    def _patch_json(self, api_url, timeout, json_data):
        """Sends a PATCH request to the given URL.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param json_data: Pre-formatted JSON to send
        :return: The JSON response as a dict, or None if none was sent
        """
        resp = self._patch(api_url,
                           timeout,
                           data=json_data,
                           content_type="application/json")

        if resp.content:
            return json.loads(resp.content)
        return None

    def _get(self, api_url, timeout, params=None) -> Response:
        """Send a GET request to the given URL, managing authentication and
        error handling, if necessary.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param params: The "query_string" to add to the url. In the format
            of a dict, {"key": "value", ...} key and val must be a string
        :return: The response object
        """
        request = requests.Request(
            method="GET",
            url=self._full_url(api_url),
            params=params)

        return self._send_authorized(request, timeout)

    def _put(self, api_url,
             timeout,
             data: Union[bytes, str],
             content_type: str) \
            -> Response:
        """Send a PUT request to the given URL, managing authentication and
        error handling, if necessary.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param data: The data to send
        :param content_type: The content type of the data
        :return: The response object
        """
        headers = None
        if content_type is not None:
            headers = {"content-type": content_type}

        request = requests.Request(
            method="PUT",
            url=self._full_url(api_url),
            data=data,
            headers=headers)

        return self._send_authorized(request, timeout)

    def _post(self, api_url,
              timeout,
              data: Union[bytes, str, BinaryIO] = None,
              content_type: str = None,
              files=None) \
            -> Response:
        """Send a POST request to the given URL, managing authentication and
        error handling, if necessary.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param data: The data to send
        :param content_type: The content type of the data
        :param files: If provided, the POST request will be a multipart request
        :return: The response object
        """
        headers = None
        if content_type is not None:
            headers = {"content-type": content_type}

        request = requests.Request(
            method="POST",
            url=self._full_url(api_url),
            data=data,
            files=files,
            headers=headers)

        return self._send_authorized(request, timeout)

    def _delete(self, api_url, timeout, params=None) -> Response:
        """Sends a DELETE request to the given URL, managing authentication and
        handling errors, as necessary.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param params: The "query_string" to add to the URL. Must be a
          str-to-str dict
        :return: The response object
        """
        request = requests.Request(
            method="DELETE",
            url=self._full_url(api_url),
            params=params)

        return self._send_authorized(request, timeout)

    def _patch(self, api_url,
               timeout,
               data: Union[bytes, str] = None,
               content_type: str = None) -> Response:
        """Sends a PATCH request to the given URL, managing authentication and
        handling errors. as necessary.

        :param api_url: The /api/blah/blah to append to the base_url
        :param timeout: The timeout to use for this request
        :param data: The data to send
        :param content_type: The content type of the data
        :return: The response object
        """
        headers = None
        if content_type is not None:
            headers = {"content-type": content_type}

        request = requests.Request(
            method="PATCH",
            url=self._full_url(api_url),
            data=data,
            headers=headers)

        return self._send_authorized(request, timeout)

    def _full_url(self, api_url):
        """Converts an API URL path to a fully qualified URL.

        :param api_url: A URL path in the format /api/blah/blah
        :return: Full URL
        """
        url = "{base_url}{api_url}".format(
            base_url=self._server_url,
            api_url=api_url)
        return url

    def _send_authorized(self, request: requests.Request, timeout) \
            -> Response:
        """Sends the given request, using whatever authorization path that is
        necessary and raising any errors.
        """
        if self._credentials is None:
            # No credentials provided, send the request without any auth
            resp = self._send_no_auth(request, timeout)
        elif self._session_id is None:
            # Authenticate with username and password to get a new session ID
            resp = self._send_with_credentials(request, timeout)
        else:
            # Authenticate with the session ID
            resp = self._send_with_session_id(request, timeout)

        return resp

    def _send_no_auth(self, request: requests.Request, timeout) \
            -> requests.Response:
        """Sends the given request with no authorization."""
        resp = self._send_request(request, timeout)
        if not resp.ok:
            raise _make_api_error(resp.content, resp.status_code)

        return resp

    def _send_with_credentials(self, request: requests.Request, timeout) \
            -> requests.Response:
        """Sends the given request with HTTP Basic Authorization."""
        request.auth = self._credentials

        resp = self._send_request(request, timeout)
        if not resp.ok:
            raise _make_api_error(resp.content, resp.status_code)

        if "session_id" in resp.cookies:
            # Update the session ID if we don't already have one
            self._session_id = resp.cookies["session_id"]

        return resp

    def _send_with_session_id(self, request: requests.Request, timeout) \
            -> requests.Response:
        """Sends the given request with the session ID."""
        request.cookies = {"session_id": self._session_id}

        try:
            resp = self._send_request(request, timeout)
            if not resp.ok:
                raise _make_api_error(resp.content, resp.status_code)
        except InvalidSessionError:
            # The session likely expired. Try again with the username and
            # password to fetch a new session
            request.cookies = None
            return self._send_with_credentials(request, timeout)

        return resp

    @staticmethod
    def _send_request(request: requests.Request, timeout: int) \
            -> requests.Response:
        """Sends a request to the server. This method is mocked out in unit
        tests.

        :param request: The request to send
        :return: The response data
        """
        prepared = request.prepare()
        return requests.Session().send(prepared, stream=True, timeout=timeout)


def _make_api_error(resp_content, status_code):
    """Makes the corresponding error for this response.

    :param resp_content: The HTTP response to inspect for info
    :return: An error that can be thrown describing this failure
    """
    if len(resp_content) == 0:
        kind = UnknownError.kind
        description = ("A failure happened but the server did not respond "
                       "with a proper error")
    else:
        try:
            resp_content = json.loads(resp_content)
            kind = resp_content["title"]
            description = resp_content["description"]
        except ValueError:
            # The content of the error was not in the proper format. This might
            # happen if some part of our request handling pipeline failed that
            # doesn't know about our error handling format. Not ideal.
            kind = UnknownError.kind
            resp_content = resp_content.decode("utf-8")
            description = ("A failure happened, and the response was not in "
                           "the proper error format: " + resp_content)

    if kind not in kind_to_error_type:
        info = f"Unknown error kind {kind}: " + description
        logging.error(info)
        return UnknownError(info, status_code)
    else:
        if kind == UnknownError.kind:
            return kind_to_error_type[kind](description, status_code)
        return kind_to_error_type[kind](description)



from typing import List, Optional, Tuple

import json

from brainframe.api.bf_codecs import Encoding, Identity, SortOptions
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class IdentityStubMixin(BaseStub):
    """Provides stubs to call APIs that create and update identities, as well
    as add new examples of the identity in image or vector form.
    """

    def get_identity(self, identity_id: int,
                     timeout=DEFAULT_TIMEOUT) -> Identity:
        """Gets the identity with the given ID.

        :param identity_id: The ID of the identity to get
        :param timeout: The timeout to use for this request
        :return: Identity
        """
        req = f"/api/identities/{identity_id}"
        identity, _ = self._get_json(req, timeout)

        return Identity.from_dict(identity)

    def get_identities(self, unique_name: str = None,
                       encoded_for_class: str = None,
                       search: Optional[str] = None,
                       limit: int = None,
                       offset: int = None,
                       sort_by: SortOptions = None,
                       timeout=DEFAULT_TIMEOUT) \
            -> Tuple[List[Identity], int]:
        """Returns all identities from the server.

        :param unique_name: If provided, identities will be filtered by only
            those who have the given unique name
        :param encoded_for_class: If provided, identities will be filtered for
            only those that have been encoded at least once for the given class
        :param search: If provided, only identities that in some way contain
            the given search query are returned. This is intended for UI search
            features, and as such the specific semantics of how the search is
            performed are subject to change.
        :param limit: If provided, the number of returned identities is limited
            to this value.
        :param offset: The offset to start limiting results from. This is only
            useful when providing a limit.
        :param sort_by: If provided, the results will be sorted by the given
            configuration
        :param timeout: The timeout to use for this request
        :return: A list of identities, and the total number of identities that
            fit this criteria, ignoring pagination (the limit and offset)
        """
        req = f"/api/identities"

        params = {}
        if unique_name is not None:
            params["unique_name"] = unique_name
        if encoded_for_class is not None:
            params["encoded_for_class"] = encoded_for_class
        if search is not None:
            params["search"] = search
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if sort_by is not None:
            params["sort_by"] = sort_by.query_format()

        identities, headers = self._get_json(req, timeout, params=params)
        identities = [Identity.from_dict(d) for d in identities]

        total_count = int(headers["Total-Count"])

        return identities, total_count

    def set_identity(self, identity: Identity,
                     timeout=DEFAULT_TIMEOUT) -> Identity:
        """Updates or creates an identity. If the identity does not already
        exist, identity.id must be None. The returned identity will have an
        assigned ID.

        :param identity: The identity to save or create
        :param timeout: The timeout to use for this request
        :return: the saved identity
        """
        req = f"/api/identities"
        saved = self._post_codec(req, timeout, identity)
        return Identity.from_dict(saved)

    def delete_identity(self, identity_id: int,
                        timeout=DEFAULT_TIMEOUT):
        """Deletes the identity with the given ID.

        :param identity_id: The ID of the identity to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/identities/{identity_id}"
        self._delete(req, timeout)

    def new_identity_image(self, identity_id: int, class_name: str,
                           storage_id: int,
                           timeout=DEFAULT_TIMEOUT):
        """Saves and encodes an image under the identity with the given ID.

        :param identity_id: Identity to associate the image with
        :param class_name: The type of object this image shows and should be
            encoded for
        :param storage_id: The ID of the image in storage to encode
        :param timeout: The timeout to use for this request
        """
        req = f"/api/identities/{identity_id}/images"
        req_obj = {
            "class_name": class_name,
            "storage_id": storage_id
        }
        encoding = self._post_json(req, timeout, json.dumps(req_obj))
        return Encoding.from_dict(encoding)

    def new_identity_vector(self, identity_id: int, class_name: str,
                            vector: List[float],
                            timeout=DEFAULT_TIMEOUT) -> int:
        """Saves the given vector under the identity with the given ID. In this
        case, a vector is simply a list of one or more numbers that describe
        some object in an image.

        :param identity_id: Identity to associate the vector with
        :param class_name: The type of object this vector describes
        :param vector: The vector to save
        :param timeout: The timeout to use for this request
        :return: The vector ID
        """
        req = f"/api/identities/{identity_id}/vectors"

        encoded_obj = {
            "class_name": class_name,
            "vector": vector
        }
        encoding = self._post_json(req, timeout, json.dumps(encoded_obj))
        return Encoding.from_dict(encoding)



from typing import List

from brainframe.api.bf_codecs import Premises
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class PremisesStubMixin(BaseStub):
    """Provides stubs for calling APIs to get, set, and delete zones."""

    def get_all_premises(self,
                         timeout=DEFAULT_TIMEOUT) -> List[Premises]:
        """Gets all premises."""
        req = "/api/premises"
        data, _ = self._get_json(req, timeout)
        zones = [Premises.from_dict(j) for j in data]
        return zones

    def get_premises(self, premises_id: int,
                     timeout=DEFAULT_TIMEOUT) -> Premises:
        """Gets the premises with the given ID.

        :param premises_id: The ID of the premises to get
        :param timeout: The timeout to use for this request
        :return: The premises
        """
        req = f"/api/premises/{premises_id}"
        data, _ = self._get_json(req, timeout)

        return Premises.from_dict(data)

    def set_premises(self, premises: Premises,
                     timeout=DEFAULT_TIMEOUT):
        """Update or create a premises. If the Premises doesn't exist, the
        premises.id must be None. An initialized Premises with an ID will be
        returned.

        Premises are more often created using the StreamGateway client tool.

        :param premises: A Premises object
        :param timeout: The timeout to use for this request
        :return: Premises, initialized with an ID
        """

        req = "/api/premises"
        data = self._post_codec(req, timeout, premises)
        new_premises = Premises.from_dict(data)
        return new_premises

    def delete_premises(self, premises_id: int,
                        timeout=DEFAULT_TIMEOUT):
        """Delete a premises and the streams and data connected to it.

        :param premises_id: The ID of the premises to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/premises/{premises_id}"
        self._delete(req, timeout)



from typing import Dict, List, Any
from enum import Enum

from dataclasses import dataclass

from .base_codecs import Codec


class OptionType(Enum):
    """The data type of a plugin option"""

    FLOAT = "float"
    INT = "int"
    ENUM = "enum"
    BOOL = "bool"

    @classmethod
    def values(cls):
        return [v.value for v in cls]


@dataclass
class PluginOption(Codec):
    """A single configuration option for a plugin. Defines what type of option
    it is and its potential values.

    There are two kinds of plugin options. Stream plugin options apply only to
    the stream they are attached to. Global plugin options apply to all
    streams, but are overridden by stream plugin options.
    """

    type: OptionType
    """The data type of this option's value"""

    default: Any
    """The default value for this option"""

    constraints: dict
    """Describes the range of valid values for this option. The content of
    this dict depends on the type field.
    
    OptionType.FLOAT:
        ``max_val``: The maximum valid float value
        
        ``min_val``: The minimum valid float value
    
    OptionType.INT:
        ``max_val``: The maximum valid int value
        
        ``min_val``: The minimum valid int value
    
    OptionType.ENUM:
        ``choices``: A list of strings. The option's value must be one of
        these strings.
    
    OptionType.BOOL:
        This object has no constraints.
    """

    description: str
    """A human-readable description of the plugin's capabilities"""

    def to_dict(self):
        d = dict(self.__dict__)
        d["type"] = self.type.value
        return d

    @staticmethod
    def from_dict(d):
        type_ = OptionType(d["type"])
        return PluginOption(type=type_,
                            default=d["default"],
                            constraints=d["constraints"],
                            description=d["description"])


class SizeType(Enum):
    """Describes the amount of DetectionNodes a plugin takes as input or
    provides as output.
    """

    NONE = "none"
    """Input: The plugin takes nothing as input, like for an object detector.
    
    Output: Plugins cannot have a NONE output.
    """
    SINGLE = "single"
    """Input: The plugin takes a single DetectionNode as input, like for a
    classifier.
    
    Output: The plugin provides a single modified DetectionNode as output, like
    for a classifier.
    """
    ALL = "all"
    """Input: The plugin takes all instances of a class as input, like for a
    tracker.
    
    Output: The plugin provides all instances of a class as output, like for a
    detector.
    """

    @classmethod
    def values(cls):
        return [v.value for v in cls]


@dataclass
class NodeDescription(Codec):
    """A description of a DetectionNode, used by plugins to define what kinds
    of inputs and outputs a plugin uses.
    """

    size: SizeType
    """Describes the amount of DetectionNodes the node either takes in as
    input or provides as output
    """

    detections: List[str]
    """A list of detection class names, like “person” or “vehicle”. A
    DetectionNode that meets this description must have a class name that
    is present in this list.
    """

    attributes: Dict[str, List[str]]
    """Key-value pairs whose key is the classification type and whose value
    is a list of possible attributes. A DetectionNode that meets this
    description must have a classification for each classification type
    listed here.
    """

    encoded: bool
    """If True, the DetectionNode must be encoded to meet this description
    """

    tracked: bool
    """If True, the DetectionNode must be tracked to meet this description
    """

    extra_data: List[str]
    """A list of keys in a NodeDescription's extra_data. A DetectionNode
    that meets this description must have extra data for each name listed
    here.
    """

    def to_dict(self):
        d = dict(self.__dict__)
        d["size"] = self.size.value
        return d

    @staticmethod
    def from_dict(d):
        size = SizeType(d["size"])
        return NodeDescription(
            size=size,
            detections=d["detections"],
            attributes=d["attributes"],
            encoded=d["encoded"],
            tracked=d["tracked"],
            extra_data=d["extra_data"])


@dataclass
class Plugin(Codec):
    """Metadata on a loaded plugin."""

    name: str
    """The name of the plugin"""

    version: int
    """The plugin's version"""

    description: str
    """A human-readable description of what the plugin does"""

    input_type: NodeDescription
    """Describes the type of inference data that this plugin takes as input
    """

    output_type: NodeDescription
    """Describes the type of inference data that this plugin produces"""

    capability: NodeDescription
    """A NodeDescription which describes what this plugin does to its
    input. It is the difference between the input and output
    NodeDescriptions. This field is useful for inspecting a plugin to find
    what it can do.
    """

    options: Dict[str, PluginOption]
    """A dict describing the configurable options of this plugin"""

    def to_dict(self):
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "input_type": self.input_type.to_dict(),
            "output_type": self.output_type.to_dict(),
            "capability": self.capability.to_dict(),
            "options": {key: val.to_dict()
                        for key, val in self.options.items()},
        }

    @staticmethod
    def from_dict(d):
        return Plugin(
            name=d["name"],
            version=d["version"],
            description=d["description"],
            input_type=NodeDescription.from_dict(d["input_type"]),
            output_type=NodeDescription.from_dict(d["output_type"]),
            capability=NodeDescription.from_dict(d["capability"]),
            options={key: PluginOption.from_dict(val)
                     for key, val in d["options"].items()},
        )



from typing import List, Optional

from dataclasses import dataclass, field

from .base_codecs import Codec


@dataclass
class Identity(Codec):
    """A specific, recognizable object or person."""

    unique_name: str
    """The unique id of the identified detection.

    Not to be confused with the id of the object which is a field used by
    the database.
    """

    nickname: str
    """A display name for the identity which may not be unique, like a
    person's name.
    """

    metadata: dict = field(default_factory=dict)
    """Any additional user-defined information about the identity."""

    id: Optional[int] = None
    """A unique identifier."""

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_dict(d):
        return Identity(
            id=d["id"],
            unique_name=d["unique_name"],
            nickname=d["nickname"],
            metadata=d["metadata"])


@dataclass
class Encoding(Codec):
    """An encoding attached to an identity."""

    identity_id: int
    """The ID of the identity this encoding is associated with."""

    class_name: str
    """The class of object this encoding is for."""

    from_image: Optional[int]
    """The storage ID of the image that this encoding was created from, or
    None if this encoding was not created from an image.
    """

    vector: List[int]
    """A low-dimensional representation of the object's appearance. This is
    what objects found in streams will be compared to in order to decide if
    the object is of the identity this encoding is associated with.
    """

    id: Optional[int] = None
    """The unique ID of the encoding."""

    def to_dict(self):
        d = dict(self.__dict__)
        return d

    @staticmethod
    def from_dict(d):
        return Encoding(id=d["id"],
                        identity_id=d["identity_id"],
                        class_name=d["class_name"],
                        from_image=d["from_image"],
                        vector=d["vector"])



import logging
from threading import RLock, Thread
from time import sleep
from typing import Any, Callable, Dict, List, Optional

import requests

from .bf_errors import UnknownError
from .bf_codecs import ZoneStatus
from .stubs.zone_statuses import ZONE_STATUS_STREAM_TYPE, ZONE_STATUS_TYPE


class StatusReceiver:
    """A simple wrapper around API.get_zone_status_stream that handles reading
    from the stream in a separate thread and notifies listeners when new data
    is received.

    The StatusReceiver corresponding to an API object can be retrieved using
    get_status_receiver. This is usually preferred over creating a
    StatusReceiver manually.
    """

    def __init__(self, api):
        """Creates a new StatusReceiver. Consider using the StatusReceiver that
        comes included with the API object with get_status_receiver instead.

        :param api: Used to communicate with the BrainFrame server
        """
        self._api = api

        self._listeners: List[Callable[[ZONE_STATUS_TYPE], Any]] = []
        self._listener_lock = RLock()

        self._latest_statuses = {}

        # no cleanup code so we can just die when the main thread dies
        self.daemon = True

        self._thread = Thread(name="StatusReceiverThread",
                              target=self._run)
        self._running = True
        self._thread.start()

    def add_listener(self, listener: Callable[[ZONE_STATUS_TYPE], Any]):
        """
        :param listener: Called when new data is received from the zone status
            stream
        """
        with self._listener_lock:
            self._listeners.append(listener)

    @property
    def is_running(self) -> bool:
        return self._running

    def latest_statuses(self, stream_id: int) -> Dict[str, ZoneStatus]:
        """Returns the latest cached list of ZoneStatuses for that stream_id,
        or any empty dict if none are cached"""
        return self._latest_statuses.get(stream_id, {})

    def close(self) -> None:
        """Close the status receiving thread"""
        self._running = False
        self._thread.join()

    def _ingest_zone_statuses(self, zone_statuses: ZONE_STATUS_TYPE):
        self._latest_statuses = zone_statuses

        with self._listener_lock:
            for listener in self._listeners:
                listener(zone_statuses)

    def _run(self):
        """Opens a connection with BrainFrame to receive ZoneStatus objects.
        Then, alerts any event handlers of new objects.
        """
        self._running = True
        zone_status_stream: Optional[ZONE_STATUS_STREAM_TYPE] = None

        while self._running:
            if zone_status_stream is None:
                zone_status_stream = self._api.get_zone_status_stream(
                    timeout=self._ZONE_STATUS_STREAM_TIMEOUT)
            try:
                zone_statuses = next(zone_status_stream)
            except (StopIteration,
                    requests.exceptions.RequestException,
                    UnknownError) as ex:

                # Catch any 502 errors that happen during server restart
                if isinstance(ex, UnknownError) \
                        and ex.status_code != 502:
                    raise

                if not self._running:
                    break

                zone_status_stream = None

                logging.warning(f"StatusLogger: Could not reach server: {ex}")
                sleep(1)
            else:
                self._ingest_zone_statuses(zone_statuses)

        self._running = False

    _ZONE_STATUS_STREAM_TIMEOUT = 10
    """The timeout in seconds before reinitializing a connection to the server. 
    If none, the zone status stream will wait forever even if there are no 
    results, making the thread potentially never check if it should be 
    closing.
    """



import json

from brainframe.api.bf_codecs import LicenseInfo
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class LicenseStubMixIn(BaseStub):
    """Provides stubs to call APIs for uploading license keys and getting
    the server's license status.
    """

    def get_license_info(self, timeout=DEFAULT_TIMEOUT) -> LicenseInfo:
        """Gets licensing information from the server.

        :param timeout: The timeout to use for this request
        :return: License info
        """
        req = "/api/license"

        license_info, _ = self._get_json(req, timeout)
        return LicenseInfo.from_dict(license_info)

    def set_license_key(self, license_key: str, timeout=DEFAULT_TIMEOUT) \
            -> LicenseInfo:
        """Uploads a license key to the server and applies it.

        :param license_key: A base64-encoded string of license data
        :param timeout: The timeout to use for this request
        :return: License info after the key is applied
        """
        req = "/api/license"

        resp = self._put(req, timeout,
                         data=license_key,
                         content_type="application/base64")
        license_info = json.loads(resp.content)
        return LicenseInfo.from_dict(license_info)



from setuptools import setup, find_namespace_packages

setup(
    name="brainframe-api",
    description="Provides a Python wrapper around the BrainFrame REST API.",
    long_description=open("README.rst").read(),
    author="Aotu",
    url="https://github.com/aotuai/brainframe_python",
    license="BSD-3-Clause",

    setup_requires=[
        "setuptools_scm",
    ],
    use_scm_version={
        # We want to be able to push these releases to PyPI, which doesn't
        # support local versions. Local versions are anything after the "+" in
        # a version string like "0.26.0.dev16+cooltext".
        "local_scheme": "no-local-version",
    },

    packages=find_namespace_packages(
        include=["brainframe.api*"]
    ),

    install_requires=[
        "requests==2.*",
        "pillow==6.*",
        "numpy==1.*",
        "dataclasses==0.*",
    ],
)
